from flask import current_app
from flask.views import View

from .utils import LazyView


class LazyViews(object):
    """
    Main instance for adding *lazy* views to Flask application or blueprint.
    """
    import_prefix = None
    instance = None

    def __init__(self, instance=None, import_prefix=None):
        """
        Initialize ``LazyViews`` instance.

        Basically it requires ``app`` or ``blueprint`` instance as first
        argument, but you could leave it empty and initialize it later with
        manually call ``init_app`` method. It could be helpful, if you want to
        configure ``LazyViews`` instance somewhere outside your ``app.py`` or
        for multiple applications.
        """
        if instance:
            self.init_app(instance, import_prefix)

    def add(self, url_rule, mixed, **options):
        """
        Add URL rule to Flask application or blueprint.

        ``mixed`` could be a real callable function, or a string Python path
        to callable view function. If ``mixed`` is a string, it would be
        wrapped with ``LazyView`` class.
        """
        assert self.instance, 'LazyViews instance is not properly initialized.'
        options['view_func'] = self.get_view(mixed)
        self.instance.add_url_rule(url_rule, **options)

    def add_admin(self, mixed, *args, **kwargs):
        """
        Add admin view if "Flask-Admin" extension added to application. This
        method only works for Flask applications, not blueprints.
        """
        assert self.instance, 'LazyViews instance is not properly initialized.'

        if not hasattr(self.instance, 'blueprints'):
            raise ValueError('Cannot add admin view to blueprint.')

        app = self.instance

        if not 'admin' in app.extensions:
            raise ValueError('Looks like, Flask-Admin extension not added '
                             'to current application, {0!r}'.format(app))

        admin = app.extensions['admin']
        view = self.get_view(mixed)

        if isinstance(view, LazyView):
            view = view(*args, **kwargs)

        admin.add_view(view)

    def add_error(self, code_or_exception, mixed):
        """
        Add error handler to Flask application or blueprint.
        """
        assert self.instance, 'LazyViews instance is not properly initialized.'
        self.instance.errorhandler(code_or_exception)(self.get_view(mixed))

    def add_static(self, url_rule, **options):
        """
        Add URL rule for serving static files to Flask app or blueprint.
        """
        assert self.instance, 'LazyViews instance is not properly initialized.'
        self.add(url_rule, self.instance.send_static_file, **options)

    def build_import_name(self, import_name):
        """
        Prepend import prefix to import name if it earlier defined by user.
        """
        not_empty = lambda data: filter(lambda item: item, data)
        return '.'.join(not_empty((self.import_prefix, import_name)))

    def get_view(self, mixed):
        """
        If ``mixed`` value is callable it's our view, else wrap it with
        ``LazyView`` instance.
        """
        if callable(mixed) or not isinstance(mixed, basestring):
            return mixed
        return LazyView(self.build_import_name(mixed))

    def init_app(self, app, import_prefix=None):
        """
        Configure ``LazyViews`` instance, store ``app`` or ``blueprint``
        instance and import prefix if any.
        """
        if import_prefix and import_prefix.startswith('.'):
            import_name = \
                app.import_name if app.import_name != '__main__' else ''

            assert import_name, 'You should properly configure import name ' \
                                'for {0!r} instance or edit import prefix to '\
                                'not start with ".".'.\
                                format(mixed.__class__.__name__)

            import_prefix = import_name + import_prefix

        self.import_prefix = import_prefix
        self.instance = app

    def init_blueprint(self, blueprint, import_prefix=None):
        """
        Alias for init app function, cause basically there are no important
        differences between Flask app and blueprint if we only need to add URL
        rule.
        """
        return self.init_app(blueprint, import_prefix)



from flask.views import View
from werkzeug.utils import cached_property, import_string


__all__ = ('LazyView', )


class LazyView(object):
    """
    Import view function only when necessary.
    """
    def __init__(self, name, *args, **kwargs):
        """
        Initialize ``LazyView`` instance for view that would be imported from
        ``name`` path.
        """
        self.__module__, self.__name__ = name.rsplit('.', 1)
        self.import_name = name
        self.args, self.kwargs = args, kwargs

    def __call__(self, *args, **kwargs):
        """
        Make real call to the view.
        """
        if self.args or self.kwargs:
            view = self.view(self.args, self.kwargs)
        else:
            view = self.view
        return view(*args, **kwargs)

    @cached_property
    def view(self):
        """
        Import view and cache it to current cls.
        """
        imported = import_string(self.import_name)

        if isinstance(imported, type) and issubclass(imported, View):
            view_name = self.import_name.lower().replace('view', '')
            return imported.as_view(view_name)

        return imported



#!/usr/bin/env python

import os

from distutils.core import setup


DIRNAME = os.path.dirname(__file__)

readme = open(os.path.join(DIRNAME, 'README.rst'), 'r')
README = readme.read()
readme.close()


setup(
    name='Flask-LazyViews',
    version='0.4',
    description='Registering url routes for Flask app and blueprints in ' \
                'lazy way.',
    long_description=README,
    author='Igor Davydenko',
    author_email='playpauseandstop@gmail.com',
    url='https://github.com/playpauseandstop/Flask-LazyViews',
    install_requires=[
        'Flask',
    ],
    packages=[
        'flask_lazyviews',
    ],
    platforms='any',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Operating System :: OS Independent',
        'Topic :: Utilities',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: BSD License',
    ],
    keywords='flask lazy views',
    license='BSD License',
)
